import json
import os

import graphql
from ddtrace.encoding import JSONEncoder, MsgpackEncoder
from ddtrace.ext import errors as ddtrace_errors
from ddtrace.tracer import Tracer
from ddtrace.writer import AgentWriter
from graphql import GraphQLField, GraphQLObjectType, GraphQLString
from graphql.execution import ExecutionResult
from graphql.language.parser import parse as graphql_parse
from graphql.language.source import Source as GraphQLSource
from wrapt import FunctionWrapper

import ddtrace_graphql
from ddtrace_graphql import (
    DATA_EMPTY, ERRORS, INVALID, QUERY, SERVICE, CLIENT_ERROR,
    TracedGraphQLSchema, patch, traced_graphql, unpatch
)
from ddtrace_graphql.base import traced_graphql_wrapped


class DummyWriter(AgentWriter):
    """
    # NB: This is coy fo DummyWriter class from ddtraces tests suite

    DummyWriter is a small fake writer used for tests. not thread-safe.
    """

    def __init__(self):
        # original call
        super(DummyWriter, self).__init__()
        # dummy components
        self.spans = []
        self.traces = []
        self.services = {}
        self.json_encoder = JSONEncoder()
        self.msgpack_encoder = MsgpackEncoder()

    def write(self, spans=None, services=None):
        if spans:
            # the traces encoding expect a list of traces so we
            # put spans in a list like we do in the real execution path
            # with both encoders
            trace = [spans]
            self.json_encoder.encode_traces(trace)
            self.msgpack_encoder.encode_traces(trace)
            self.spans += spans
            self.traces += trace

        if services:
            self.json_encoder.encode_services(services)
            self.msgpack_encoder.encode_services(services)
            self.services.update(services)

    def pop(self):
        # dummy method
        s = self.spans
        self.spans = []
        return s

    def pop_traces(self):
        # dummy method
        traces = self.traces
        self.traces = []
        return traces

    def pop_services(self):
        # dummy method
        s = self.services
        self.services = {}
        return s


def get_dummy_tracer():
    tracer = Tracer()
    tracer.writer = DummyWriter()
    return tracer


def get_traced_schema(tracer=None, query=None, resolver=None):
    resolver = resolver or (lambda *_: 'world')
    tracer = tracer or get_dummy_tracer()
    query = query or GraphQLObjectType(
        name='RootQueryType',
        fields={
            'hello': GraphQLField(
                type=GraphQLString,
                resolver=resolver,
            )
        }
    )
    return tracer, TracedGraphQLSchema(query=query, datadog_tracer=tracer)


class TestGraphQL:

    def test_unpatch(self):
        gql = graphql.graphql
        unpatch()
        assert gql == graphql.graphql
        assert not isinstance(graphql.graphql, FunctionWrapper)
        patch()
        assert isinstance(graphql.graphql, FunctionWrapper)

        tracer, schema = get_traced_schema()
        graphql.graphql(schema, '{ hello }')
        span = tracer.writer.pop()[0]

        unpatch()
        assert gql == graphql.graphql

        cb_args = {}
        def test_cb(**kwargs):
            cb_args.update(kwargs)
        patch(span_callback=test_cb)
        assert isinstance(graphql.graphql, FunctionWrapper)

        result = graphql.graphql(schema, '{ hello }')
        span = tracer.writer.pop()[0]
        assert cb_args['span'] is span
        assert cb_args['result'] is result

        unpatch()
        assert gql == graphql.graphql



    def test_invalid(self):
        tracer, schema = get_traced_schema()
        result = traced_graphql(schema, '{ hello world }')
        span = tracer.writer.pop()[0]
        assert span.get_metric(INVALID) == result.invalid == 1
        assert span.get_metric(DATA_EMPTY) == 1
        assert span.error == 0

        result = traced_graphql(schema, '{ hello }')
        span = tracer.writer.pop()[0]
        assert span.get_metric(INVALID) == result.invalid == 0
        assert span.error == 0

    def test_unhandled_exception(self):

        def exc_resolver(*args):
            raise Exception('Testing stuff')

        tracer, schema = get_traced_schema(resolver=exc_resolver)
        result = traced_graphql(schema, '{ hello }')
        span = tracer.writer.pop()[0]
        assert span.get_metric(INVALID) == 0
        assert span.error == 1
        assert span.get_metric(CLIENT_ERROR) == 0
        assert span.get_metric(DATA_EMPTY) == 0

        error_stack = span.get_tag(ddtrace_errors.ERROR_STACK)
        assert 'Testing stuff' in error_stack
        assert 'Traceback' in error_stack

        error_msg = span.get_tag(ddtrace_errors.ERROR_MSG)
        assert 'Testing stuff' in error_msg

        error_type = span.get_tag(ddtrace_errors.ERROR_TYPE)
        assert 'Exception' in error_type

        try:
            raise Exception('Testing stuff')
        except Exception as exc:
            _error = exc

        def _tg(*args, **kwargs):
            def func(*args, **kwargs):
                return ExecutionResult(
                    errors=[_error],
                    invalid=True,
                )
            return traced_graphql_wrapped(func, args, kwargs)

        tracer, schema = get_traced_schema(resolver=exc_resolver)
        result = _tg(schema, '{ hello }')

        span = tracer.writer.pop()[0]
        assert span.get_metric(INVALID) == 1
        assert span.error == 1
        assert span.get_metric(DATA_EMPTY) == 1

        error_stack = span.get_tag(ddtrace_errors.ERROR_STACK)
        assert 'Testing stuff' in error_stack
        assert 'Traceback' in error_stack

    def test_not_server_error(self):
        class TestException(Exception):
            pass

        def exc_resolver(*args):
            raise TestException('Testing stuff')

        tracer, schema = get_traced_schema(resolver=exc_resolver)
        result = traced_graphql(
            schema,
            '{ hello }',
            ignore_exceptions=(TestException),
        )
        span = tracer.writer.pop()[0]
        assert span.get_metric(INVALID) == 0
        assert span.error == 0
        assert span.get_metric(DATA_EMPTY) == 0
        assert span.get_metric(CLIENT_ERROR) == 1

    def test_request_string_resolve(self):
        query = '{ hello }'

        # string as args[1]
        tracer, schema = get_traced_schema()
        traced_graphql(schema, query)
        span = tracer.writer.pop()[0]
        assert span.get_tag(QUERY) == query

        # string as kwargs.get('request_string')
        tracer, schema = get_traced_schema()
        traced_graphql(schema, request_string=query)
        span = tracer.writer.pop()[0]
        assert span.get_tag(QUERY) == query

        # ast as args[1]
        tracer, schema = get_traced_schema()
        ast_query = graphql_parse(GraphQLSource(query, 'Test Request'))
        traced_graphql(schema, ast_query)
        span = tracer.writer.pop()[0]
        assert span.get_tag(QUERY) == query

        # ast as kwargs.get('request_string')
        tracer, schema = get_traced_schema()
        ast_query = graphql_parse(GraphQLSource(query, 'Test Request'))
        traced_graphql(schema, request_string=ast_query)
        span = tracer.writer.pop()[0]
        assert span.get_tag(QUERY) == query

    @staticmethod
    def test_query_tag():
        query = '{ hello }'
        tracer, schema = get_traced_schema()
        traced_graphql(schema, query)
        span = tracer.writer.pop()[0]
        assert span.get_tag(QUERY) == query

        # test query also for error span, just in case
        query = '{ hello world }'
        tracer, schema = get_traced_schema()
        traced_graphql(schema, query)
        span = tracer.writer.pop()[0]
        assert span.get_tag(QUERY) == query

    @staticmethod
    def test_errors_tag():
        query = '{ hello }'
        tracer, schema = get_traced_schema()
        result = traced_graphql(schema, query)
        span = tracer.writer.pop()[0]
        assert not span.get_tag(ERRORS)
        assert result.errors is span.get_tag(ERRORS) is None

        query = '{ hello world }'
        result = traced_graphql(schema, query)
        span = tracer.writer.pop()[0]
        span_errors = span.get_tag(ERRORS)
        assert span_errors
        _se = json.loads(span_errors)
        assert len(_se) == len(result.errors) == 1
        assert 'message' in _se[0]
        assert 'line' in _se[0]['locations'][0]
        assert 'column' in _se[0]['locations'][0]

    @staticmethod
    def test_resource():
        query = '{ hello world }'
        tracer, schema = get_traced_schema()
        traced_graphql(schema, query)
        span = tracer.writer.pop()[0]
        assert span.resource == query

        query = 'mutation fnCall(args: Args) { }'
        traced_graphql(schema, query)
        span = tracer.writer.pop()[0]
        assert span.resource == 'mutation fnCall'

        query = 'mutation fnCall { }'
        traced_graphql(schema, query)
        span = tracer.writer.pop()[0]
        assert span.resource == 'mutation fnCall'

        query = 'mutation fnCall { }'
        traced_graphql(schema, query, span_kwargs={'resource': 'test'})
        span = tracer.writer.pop()[0]
        assert span.resource == 'test'

    @staticmethod
    def test_span_callback():
        cb_args = {}
        def test_cb(result, span):
            cb_args.update(dict(result=result, span=span))
        query = '{ hello world }'
        tracer, schema = get_traced_schema()
        result = traced_graphql(schema, query, span_callback=test_cb)
        span = tracer.writer.pop()[0]
        assert cb_args['span'] is span
        assert cb_args['result'] is result

    @staticmethod
    def test_span_kwargs_overrides():
        query = '{ hello }'
        tracer, schema = get_traced_schema()

        traced_graphql(schema, query, span_kwargs={'resource': 'test'})
        span = tracer.writer.pop()[0]
        assert span.resource == 'test'

        traced_graphql(
            schema,
            query,
            span_kwargs={
                'service': 'test',
                'name': 'test',
            }
        )
        span = tracer.writer.pop()[0]
        assert span.service == 'test'
        assert span.name == 'test'
        assert span.resource == '{ hello }'

    @staticmethod
    def test_service_from_env():
        query = '{ hello }'
        tracer, schema = get_traced_schema()

        global traced_graphql
        traced_graphql(schema, query)
        span = tracer.writer.pop()[0]
        assert span.service == SERVICE

        os.environ['DDTRACE_GRAPHQL_SERVICE'] = 'test.test'

        traced_graphql(schema, query)
        span = tracer.writer.pop()[0]
        assert span.service == 'test.test'

    @staticmethod
    def test_tracer_disabled():
        query = '{ hello world }'
        tracer, schema = get_traced_schema()
        tracer.enabled = False
        traced_graphql(schema, query)
        assert not tracer.writer.pop()



import json
import re
import traceback
from io import StringIO

from graphql.error import GraphQLError, format_error
from graphql.language.ast import Document


def get_request_string(args, kwargs):
    """
    Given ``args``, ``kwargs`` of original function, returns request string.
    """
    return args[1] if len(args) > 1 else kwargs.get('request_string')


def get_query_string(args, kwargs):
    """
    Given ``args``, ``kwargs`` of original function, returns query as string.
    """
    rs = get_request_string(args, kwargs)
    return rs.loc.source.body if isinstance(rs, Document) else rs


def is_server_error(result, ignore_exceptions):
    """
    Determines from ``result`` if server error occured.

    Based on error handling done here https://bit.ly/2JamxWF
    """
    errors = None if result.errors is None else [
        error for error in result.errors
        if not isinstance(original_error(error), ignore_exceptions)
    ]
    return bool(
        (
            errors
            and not result.invalid
        )
        or
        (
            errors
            and result.invalid
            and len(result.errors) == 1
            and not isinstance(result.errors[0], GraphQLError)
        )
    )


def original_error(err):
    """
    Returns original exception from graphql exception wrappers.

    graphql-core wraps exceptions that occurs on resolvers into special type
    with ``original_error`` attribute, which contains the real exception.
    """
    return err.original_error if hasattr(err, 'original_error') else err


def format_errors(errors):
    """
    Formats list of exceptions, ``errors``, into json-string.

    ``GraphQLError exceptions`` contain additional information like line and
    column number, where the exception at query resolution happened. This
    method tries to extract that information.
    """
    return json.dumps(
        [
            # fix for graphql-core==1.x
            format_error(err) if hasattr(err, 'message') else str(err)
            for err in errors
        ],
        indent=2,
    )


def format_error_traceback(error, limit=20):
    """
    Returns ``limit`` lines of ``error``s exception traceback.
    """
    buffer_file = StringIO()
    traceback.print_exception(
        type(error),
        error,
        error.__traceback__,
        file=buffer_file,
        limit=limit,
    )
    return buffer_file.getvalue()


def format_errors_traceback(errors):
    """
    Concatenates traceback strings from list of exceptions in ``errors``.
    """
    return "\n\n".join([
        format_error_traceback(original_error(error))
        for error in errors if isinstance(error, Exception)
    ])


def _err_msg(error):
    return str(original_error(error))


def format_errors_msg(errors):
    """
    Formats error message as json string from list of exceptions ``errors``.
    """
    return _err_msg(errors[0]) if len(errors) == 1 else json.dumps(
        [
            _err_msg(error)
            for error in errors if isinstance(error, Exception)
        ],
        indent=2
    )


def _err_type(error):
    return type(original_error(error)).__name__


def format_errors_type(errors):
    """
    Formats error types as json string from list of exceptions ``errors``.
    """
    return _err_type(errors[0]) if len(errors) == 1 else json.dumps(
        [
            _err_type(error)
            for error in errors if isinstance(error, Exception)
        ],
        indent=2
    )


def resolve_query_res(query):
    """
    Extracts resource name from ``query`` string.
    """
    # split by '(' for queries with arguments
    # split by '{' for queries without arguments
    # rather full query than empty resource name
    return re.split('[({]', query, 1)[0].strip() or query



"""
Tracing for the graphql-core library.

https://github.com/graphql-python/graphql-core
"""

import logging
import os

import graphql
import graphql.backend.core
import wrapt
from ddtrace.internal.wrapping import unwrap

from ddtrace_graphql.base import traced_graphql_wrapped

logger = logging.getLogger(__name__)


def patch(span_kwargs=None, span_callback=None, ignore_exceptions=()):
    """
    Monkeypatches graphql-core library to trace graphql calls execution.
    """

    def wrapper(func, _, args, kwargs):
        return traced_graphql_wrapped(
            func,
            args,
            kwargs,
            span_kwargs=span_kwargs,
            span_callback=span_callback,
            ignore_exceptions=ignore_exceptions,
        )

    logger.debug("Patching `graphql.graphql` function.")

    wrapt.wrap_function_wrapper(graphql, "graphql", wrapper)

    logger.debug("Patching `graphql.backend.core.execute_and_validate` function.")

    wrapt.wrap_function_wrapper(graphql.backend.core, "execute_and_validate", wrapper)


def unpatch():
    logger.debug("Unpatching `graphql.graphql` function.")
    unwrap(graphql, "graphql")
    unwrap(graphql.backend.core, "execute_and_validate")



"""
To trace all GraphQL requests, patch the library like so::

    from ddtrace_graphql import patch
    patch()

    from graphql import graphql
    result = graphql(schema, query)


If you do not want to monkeypatch ``graphql.graphql`` function or want to trace
only certain calls you can use the ``traced_graphql`` function::

    from ddtrace_graphql import traced_graphql
    traced_graphql(schema, query)
"""


from ddtrace.internal.utils.importlib import require_modules

required_modules = ['graphql']

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .base import (
            TracedGraphQLSchema, traced_graphql,
            TYPE, SERVICE, QUERY, ERRORS, INVALID, RES_NAME, DATA_EMPTY,
            CLIENT_ERROR
        )
        from .patch import patch, unpatch
        __all__ = [
            'TracedGraphQLSchema',
            'patch', 'unpatch', 'traced_graphql',
            'TYPE', 'SERVICE', 'QUERY', 'ERRORS', 'INVALID',
            'RES_NAME', 'DATA_EMPTY', 'CLIENT_ERROR',
        ]




from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(here, "VERSION"), encoding="utf-8") as f:
    version = f.read()


setup(
    name="ddtrace-graphql-juaco",  # Required
    version=version,  # Required
    description="Python library for tracing graphql calls with Datadog",
    long_description=long_description,  # Optional
    long_description_content_type="text/x-rst",  # Added
    url="https://github.com/vamolin/ddtrace-graphql",  # Optional
    author="Michal Kuffa",  # Optional
    author_email="michal@bynder.com",  # Optional
    license="MIT",
    classifiers=[  # Optional
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Debuggers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    keywords="tracing datadog graphql graphene",
    packages=find_packages(exclude=["tests"]),  # Required
    install_requires=["ddtrace", "graphql-core", "wrapt"],
    extras_require={"test": ["tox", "pytest", "pytest-cov"]},
)



import logging
import os

import ddtrace
from ddtrace import constants as ddtrace_constants
import graphql

from ddtrace_graphql import utils

logger = logging.getLogger(__name__)
_graphql = graphql.graphql


TYPE = 'graphql'
QUERY = 'query'
ERRORS = 'errors'
INVALID = 'invalid'
CLIENT_ERROR = 'client_error'
DATA_EMPTY = 'data_empty'
RES_NAME = 'graphql.graphql'
#
SERVICE_ENV_VAR = 'DDTRACE_GRAPHQL_SERVICE'
SERVICE = 'graphql'


class TracedGraphQLSchema(graphql.GraphQLSchema):
    def __init__(self, *args, **kwargs):
        if 'datadog_tracer' in kwargs:
            self.datadog_tracer = kwargs.pop('datadog_tracer')
            logger.debug(
                'For schema %s using own tracer %s',
                self, self.datadog_tracer)
        super(TracedGraphQLSchema, self).__init__(*args, **kwargs)


def traced_graphql_wrapped(
    func,
    args,
    kwargs,
    span_kwargs=None,
    span_callback=None,
    ignore_exceptions=(),
):
    """
    Wrapper for graphql.graphql function.
    """
    # allow schemas their own tracer with fall-back to the global
    schema = args[0]
    tracer = getattr(schema, 'datadog_tracer', ddtrace.tracer)

    if not tracer.enabled:
        return func(*args, **kwargs)

    query = utils.get_query_string(args, kwargs)

    _span_kwargs = {
        'name': RES_NAME,
        'span_type': TYPE,
        'service': os.getenv(SERVICE_ENV_VAR, SERVICE),
        'resource': utils.resolve_query_res(query)
    }
    _span_kwargs.update(span_kwargs or {})

    with tracer.trace(**_span_kwargs) as span:
        span.set_tag(QUERY, query)
        result = None
        try:
            result = func(*args, **kwargs)
            return result
        finally:
            # `span.error` must be integer
            span.error = int(result is None)

            if result is not None:

                span.error = 0
                if result.errors:
                    span.set_tag(
                        ERRORS,
                        utils.format_errors(result.errors))
                    span.set_tag(
                        ddtrace_constants.ERROR_STACK,
                        utils.format_errors_traceback(result.errors))
                    span.set_tag(
                        ddtrace_constants.ERROR_MSG,
                        utils.format_errors_msg(result.errors))
                    span.set_tag(
                        ddtrace_constants.ERROR_TYPE,
                        utils.format_errors_type(result.errors))

                    span.error = int(utils.is_server_error(
                        result,
                        ignore_exceptions,
                    ))

                span.set_metric(
                    CLIENT_ERROR,
                    int(bool(not span.error and result.errors))
                )
                span.set_metric(INVALID, int(result.invalid))
                span.set_metric(DATA_EMPTY, int(result.data is None))

            if span_callback is not None:
                span_callback(result=result, span=span)


def traced_graphql(
    *args,
    span_kwargs=None,
    span_callback=None,
    ignore_exceptions=(),
    **kwargs
):
    return traced_graphql_wrapped(
        _graphql, args, kwargs,
        span_kwargs=span_kwargs,
        span_callback=span_callback,
        ignore_exceptions=ignore_exceptions
    )
